@@ -32,10 +32,9 @@ from ..interface import (
     AcquisitionFunctionClass,
     SingleModelAcquisitionBuilder,
 )
-from ..sampler import BatchReparametrizationSampler
 
 
-class ExpectedImprovement(SingleModelAcquisitionBuilder):
+class ExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
     """
     Builder for the expected improvement function where the "best" value is taken to be the minimum
     of the posterior mean at observed points.
@@ -122,7 +121,7 @@ class expected_improvement(AcquisitionFunctionClass):
         return (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(self._eta)
 
 
-class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder):
+class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
     """
     Builder for the augmented expected improvement function for optimization single-objective
     optimization problems with high levels of observation noise.
@@ -231,7 +230,7 @@ class augmented_expected_improvement(AcquisitionFunctionClass):
         return expected_improvement * augmentation
 
 
-class NegativeLowerConfidenceBound(SingleModelAcquisitionBuilder):
+class NegativeLowerConfidenceBound(SingleModelAcquisitionBuilder[ProbabilisticModel]):
     """
     Builder for the negative of the lower confidence bound. The lower confidence bound is typically
     minimised, so the negative is suitable for maximisation.
@@ -324,7 +323,7 @@ def lower_confidence_bound(model: ProbabilisticModel, beta: float) -> Acquisitio
     return acquisition
 
 
-class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder):
+class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]):
     r"""
     Builder for the :func:`probability_of_feasibility` acquisition function, defined in
     :cite:`gardner14` as
@@ -422,7 +421,7 @@ def probability_of_feasibility(
     return acquisition
 
 
-class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder):
+class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticModel]):
     """
     Builder for the *expected constrained improvement* acquisition function defined in
     :cite:`gardner14`. The acquisition function computes the expected improvement from the best
@@ -433,7 +432,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder):
     def __init__(
         self,
         objective_tag: str,
-        constraint_builder: AcquisitionFunctionBuilder,
+        constraint_builder: AcquisitionFunctionBuilder[ProbabilisticModel],
         min_feasibility_probability: float | TensorType = 0.5,
     ):
         """
@@ -584,11 +583,10 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder):
             self._expected_improvement_fn.update(eta)  # type: ignore
 
 
-class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder):
+class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
     """
     Expected improvement for batches of points (or :math:`q`-EI), approximated using Monte Carlo
     estimation with the reparametrization trick. See :cite:`Ginsbourger2010` for details.
-
     Improvement is measured with respect to the minimum predictive mean at observed query points.
     This is calculated in :class:`BatchMonteCarloExpectedImprovement` by assuming observations
     at new points are independent from those at known query points. This is faster, but is an
@@ -664,8 +662,9 @@ class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder):
 class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
     def __init__(self, sample_size: int, model: ProbabilisticModel, eta: TensorType, jitter: float):
         """
-
-        :param sampler:  BatchReparametrizationSampler.
+        :param sample_size: The number of Monte-Carlo samples.
+        :param model: The model of the objective function.
+        :param sampler:  ReparametrizationSampler.
         :param eta: The "best" observation.
         :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
             the covariance matrix.
@@ -674,14 +673,25 @@ class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
             greater than one.
         """
         self._sample_size = sample_size
-        self._sampler = BatchReparametrizationSampler(sample_size, model)
+
+        try:
+            sampler = model.reparam_sampler(self._sample_size)
+        except (NotImplementedError):
+            raise ValueError(
+                """
+                The batch Monte-Carlo expected improvement acquisition function
+                only supports models that implement a reparam_sampler method.
+                """
+            )
+
+        self._sampler = sampler
         self._eta = tf.Variable(eta)
         self._jitter = jitter
 
     def update(self, eta: TensorType) -> None:
-        """Update the acquisition function with a new eta value."""
+        """Update the acquisition function with a new eta value and reset the reparam sampler."""
         self._eta.assign(eta)
-        self._sampler._initialized.assign(False)
+        self._sampler.reset_sampler()
 
     @tf.function
     def __call__(self, x: TensorType) -> TensorType:
