@@ -41,7 +41,7 @@ from .interface import (
     SingleModelGreedyAcquisitionBuilder,
 )
 from .optimizer import AcquisitionOptimizer, automatic_optimizer_selector, batchify
-from .sampler import ExactThompsonSampler, RandomFourierFeatureThompsonSampler, ThompsonSampler
+from .sampler import ExactThompsonSampler, ThompsonSampler
 
 T_co = TypeVar("T_co", covariant=True)
 """ Unbound covariant type variable. """
@@ -118,10 +118,10 @@ class EfficientGlobalOptimization(AcquisitionRule[TensorType, SP_contra]):
     def __init__(
         self,
         builder: Optional[
-            AcquisitionFunctionBuilder
-            | GreedyAcquisitionFunctionBuilder
-            | SingleModelAcquisitionBuilder
-            | SingleModelGreedyAcquisitionBuilder
+            AcquisitionFunctionBuilder[ProbabilisticModel]
+            | GreedyAcquisitionFunctionBuilder[ProbabilisticModel]
+            | SingleModelAcquisitionBuilder[ProbabilisticModel]
+            | SingleModelGreedyAcquisitionBuilder[ProbabilisticModel]
         ] = None,
         optimizer: AcquisitionOptimizer[SP_contra] | None = None,
         num_query_points: int = 1,
@@ -162,7 +162,10 @@ class EfficientGlobalOptimization(AcquisitionRule[TensorType, SP_contra]):
             # Joint batch acquisitions require batch optimizers
             optimizer = batchify(optimizer, num_query_points)
 
-        self._builder: Union[AcquisitionFunctionBuilder, GreedyAcquisitionFunctionBuilder] = builder
+        self._builder: Union[
+            AcquisitionFunctionBuilder[ProbabilisticModel],
+            GreedyAcquisitionFunctionBuilder[ProbabilisticModel],
+        ] = builder
         self._optimizer = optimizer
         self._num_query_points = num_query_points
         self._acquisition_function: Optional[AcquisitionFunction] = None
@@ -348,7 +351,10 @@ class AsynchronousOptimization(
 
     def __init__(
         self,
-        builder: Optional[AcquisitionFunctionBuilder | SingleModelAcquisitionBuilder] = None,
+        builder: Optional[
+            AcquisitionFunctionBuilder[ProbabilisticModel]
+            | SingleModelAcquisitionBuilder[ProbabilisticModel]
+        ] = None,
         optimizer: AcquisitionOptimizer[SP_contra] | None = None,
         num_query_points: int = 1,
     ):
@@ -380,7 +386,7 @@ class AsynchronousOptimization(
         if num_query_points > 1:
             optimizer = batchify(optimizer, num_query_points)
 
-        self._builder: AcquisitionFunctionBuilder = builder
+        self._builder: AcquisitionFunctionBuilder[ProbabilisticModel] = builder
         self._optimizer = optimizer
         self._acquisition_function: Optional[AcquisitionFunction] = None
 
@@ -493,7 +499,8 @@ class AsynchronousGreedy(
 
     def __init__(
         self,
-        builder: GreedyAcquisitionFunctionBuilder | SingleModelGreedyAcquisitionBuilder,
+        builder: GreedyAcquisitionFunctionBuilder[ProbabilisticModel]
+        | SingleModelGreedyAcquisitionBuilder[ProbabilisticModel],
         optimizer: AcquisitionOptimizer[SP_contra] | None = None,
         num_query_points: int = 1,
     ):
@@ -528,7 +535,7 @@ class AsynchronousGreedy(
         if isinstance(builder, SingleModelGreedyAcquisitionBuilder):
             builder = builder.using(OBJECTIVE)
 
-        self._builder: GreedyAcquisitionFunctionBuilder = builder
+        self._builder: GreedyAcquisitionFunctionBuilder[ProbabilisticModel] = builder
         self._optimizer = optimizer
         self._acquisition_function: Optional[AcquisitionFunction] = None
         self._num_query_points = num_query_points
@@ -635,7 +642,8 @@ class DiscreteThompsonSampling(AcquisitionRule[TensorType, SearchSpace]):
 
     The model is sampled either exactly (with an :math:`O(N^3)` complexity), or sampled
     approximately through a random Fourier `M` feature decompisition
-    (with an :math:`O(\min(n^3,M^3))` complexity for a model trained on `n` points).
+    (with an :math:`O(\min(n^3,M^3))` complexity for a model trained on `n` points). The number
+    `M` of Fourier features is specified when building the model.
 
     """
 
@@ -643,14 +651,12 @@ class DiscreteThompsonSampling(AcquisitionRule[TensorType, SearchSpace]):
         self,
         num_search_space_samples: int,
         num_query_points: int,
-        num_fourier_features: Optional[int] = None,
+        thompson_sampler: Optional[ThompsonSampler] = None,
     ):
         """
         :param num_search_space_samples: The number of points at which to sample the posterior.
         :param num_query_points: The number of points to acquire.
-        :num_fourier_features: The number of features used to approximate the kernel. We
-            recommend first trying 1000 features, as this typically perfoms well for a wide
-            range of kernels. If None, then we perfom exact Thompson sampling.
+        :thompson_sampler: Sampler to sample maximisers from the underlying model.
         """
         if not num_search_space_samples > 0:
             raise ValueError(f"Search space must be greater than 0, got {num_search_space_samples}")
@@ -660,21 +666,27 @@ class DiscreteThompsonSampling(AcquisitionRule[TensorType, SearchSpace]):
                 f"Number of query points must be greater than 0, got {num_query_points}"
             )
 
-        if num_fourier_features is not None and num_fourier_features <= 0:
-            raise ValueError(
-                f"Number of fourier features must be greater than 0, got {num_query_points}"
-            )
+        if thompson_sampler is not None:
+            if thompson_sampler.sample_min_value:
+                raise ValueError(
+                    """
+                    Thompson sampling requires a thompson_sampler that samples minimizers,
+                    not just minimum values. However the passed sampler has sample_min_value=True.
+                    """
+                )
+        else:
+            thompson_sampler = ExactThompsonSampler(sample_min_value=False)
 
+        self._thompson_sampler = thompson_sampler
         self._num_search_space_samples = num_search_space_samples
         self._num_query_points = num_query_points
-        self._num_fourier_features = num_fourier_features
 
     def __repr__(self) -> str:
         """"""
         return f"""DiscreteThompsonSampling(
         {self._num_search_space_samples!r},
         {self._num_query_points!r},
-        {self._num_fourier_features!r})"""
+        {self._thompson_sampler!r})"""
 
     def acquire(
         self,
@@ -704,20 +716,10 @@ class DiscreteThompsonSampling(AcquisitionRule[TensorType, SearchSpace]):
                 f"""datasets must be provided and contain the single key {OBJECTIVE}"""
             )
 
-        if self._num_fourier_features is None:  # Perform exact Thompson sampling
-            thompson_sampler: ThompsonSampler = ExactThompsonSampler(
-                self._num_query_points, models[OBJECTIVE]
-            )
-        else:  # Perform approximate Thompson sampling
-            thompson_sampler = RandomFourierFeatureThompsonSampler(
-                self._num_query_points,
-                models[OBJECTIVE],
-                datasets[OBJECTIVE],
-                num_features=self._num_fourier_features,
-            )
-
         query_points = search_space.sample(self._num_search_space_samples)
-        thompson_samples = thompson_sampler.sample(query_points)
+        thompson_samples = self._thompson_sampler.sample(
+            models[OBJECTIVE], self._num_query_points, query_points
+        )
 
         return thompson_samples
 
