@@ -17,6 +17,15 @@ and its trainable subclass :class:`TrainableProbabilisticModel`. It also contain
 creating :class:`TrainableProbabilisticModel`\ s from config.
 """
 
-from . import gpflow, gpflux, optimizer
+from . import gpflow, gpflux, keras, optimizer
 from .config import ModelConfig, ModelRegistry, ModelSpec, create_model
-from .interfaces import ModelStack, ProbabilisticModel, TrainableProbabilisticModel
+from .interfaces import (
+    FastUpdateModel,
+    ModelStack,
+    ProbabilisticModel,
+    ReparametrizationSampler,
+    TrainableModelStack,
+    TrainableProbabilisticModel,
+    TrajectoryFunction,
+    TrajectorySampler,
+)
