@@ -19,26 +19,38 @@ from typing import Optional
 import gpflow
 import tensorflow as tf
 import tensorflow_probability as tfp
+from gpflow.conditionals.util import sample_mvn
 from gpflow.models import GPR, SGPR, SVGP, VGP
 from gpflow.utilities import multiple_assign, read_values
+from gpflow.utilities.ops import leading_transpose
 
 from ...data import Dataset
 from ...types import TensorType
 from ...utils import DEFAULTS, jit
-from ..interfaces import TrainableProbabilisticModel
+from ..interfaces import FastUpdateModel, TrainableProbabilisticModel, TrajectorySampler
 from ..optimizer import BatchOptimizer, Optimizer
 from .interface import GPflowPredictor
-from .utils import assert_data_is_compatible, randomize_hyperparameters, squeeze_hyperparameters
+from .sampler import RandomFourierFeatureTrajectorySampler
+from .utils import (
+    assert_data_is_compatible,
+    check_optimizer,
+    randomize_hyperparameters,
+    squeeze_hyperparameters,
+)
 
 
-class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel):
+class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel, FastUpdateModel):
     """
     A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.GPR`
     or :class:`~gpflow.models.SGPR`.
     """
 
     def __init__(
-        self, model: GPR | SGPR, optimizer: Optimizer | None = None, num_kernel_samples: int = 10
+        self,
+        model: GPR | SGPR,
+        optimizer: Optimizer | None = None,
+        num_kernel_samples: int = 10,
+        num_rff_features: int = 1000,
     ):
         """
         :param model: The GPflow model to wrap.
@@ -47,21 +59,33 @@ class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel):
         :param num_kernel_samples: Number of randomly sampled kernels (for each kernel parameter) to
             evaluate before beginning model optimization. Therefore, for a kernel with `p`
             (vector-valued) parameters, we evaluate `p * num_kernel_samples` kernels.
+        :param num_rff_features: The number of random Foruier features used to approximate the
+            kernel when calling :meth:`trajectory_sampler`. We use a default of 1000 as it
+            typically perfoms well for a wide range of kernels. Note that very smooth
+            kernels (e.g. RBF) can be well-approximated with fewer features.
         """
         super().__init__(optimizer)
         self._model = model
 
+        check_optimizer(self.optimizer)
+
         if num_kernel_samples <= 0:
             raise ValueError(
                 f"num_kernel_samples must be greater or equal to zero but got {num_kernel_samples}."
             )
         self._num_kernel_samples = num_kernel_samples
 
+        if num_rff_features <= 0:
+            raise ValueError(
+                f"num_rff_features must be greater or equal to zero but got {num_rff_features}."
+            )
+        self._num_rff_features = num_rff_features
+
         self._ensure_variable_model_data()
 
     def __repr__(self) -> str:
         """"""
-        return f"GaussianProcessRegression({self._model!r}, {self.optimizer!r})"
+        return f"GaussianProcessRegression({self.model!r}, {self.optimizer!r})"
 
     @property
     def model(self) -> GPR | SGPR:
@@ -112,34 +136,57 @@ class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel):
 
         .. math:: \Sigma_{12} = K_{12} - K_{x1}(K_{xx} + \sigma^2 I)^{-1}K_{x2}
 
-        :param query_points_1: Set of query points with shape [N, D]
-        :param query_points_2: Sets of query points with shape [M, D]
+        Note that query_points_2 must be a rank 2 tensor, but query_points_1 can
+        have leading dimensions.
 
-        :return: Covariance matrix between the sets of query points with shape [N, M]
+        :param query_points_1: Set of query points with shape [..., N, D]
+        :param query_points_2: Sets of query points with shape [M, D]
+        :return: Covariance matrix between the sets of query points with shape [..., L, N, M]
+            (L being the number of latent GPs = number of output dimensions)
         """
         if isinstance(self.model, SGPR):
             raise NotImplementedError("Covariance between points is not supported for SGPR.")
 
-        tf.debugging.assert_shapes([(query_points_1, ["N", "D"]), (query_points_2, ["M", "D"])])
+        tf.debugging.assert_shapes(
+            [(query_points_1, [..., "N", "D"]), (query_points_2, ["M", "D"])]
+        )
 
         x = self.model.data[0].value()
         num_data = tf.shape(x)[0]
         s = tf.linalg.diag(tf.fill([num_data], self.model.likelihood.variance))
 
-        K = self.model.kernel(x)
-        L = tf.linalg.cholesky(K + s)
+        K = self.model.kernel(x)  # [num_data, num_data] or [L, num_data, num_data]
+        Kx1 = self.model.kernel(query_points_1, x)  # [..., N, num_data] or [..., L, N, num_data]
+        Kx2 = self.model.kernel(x, query_points_2)  # [num_data, M] or [L, num_data, M]
+        K12 = self.model.kernel(query_points_1, query_points_2)  # [..., N, M] or [..., L, N, M]
 
-        Kx1 = self.model.kernel(x, query_points_1)
-        Linv_Kx1 = tf.linalg.triangular_solve(L, Kx1)
+        if len(tf.shape(K)) == 2:
+            K = tf.expand_dims(K, -3)
+            Kx1 = tf.expand_dims(Kx1, -3)
+            Kx2 = tf.expand_dims(Kx2, -3)
+            K12 = tf.expand_dims(K12, -3)
+        elif len(tf.shape(K)) > 3:
+            raise NotImplementedError(
+                "Covariance between points is not supported "
+                "for kernels of type "
+                f"{type(self.model.kernel)}."
+            )
 
-        Kx2 = self.model.kernel(x, query_points_2)
-        Linv_Kx2 = tf.linalg.triangular_solve(L, Kx2)
+        L = tf.linalg.cholesky(K + s)  # [L, num_data, num_data]
 
-        K12 = self.model.kernel(query_points_1, query_points_2)
-        cov = K12 - tf.tensordot(tf.transpose(Linv_Kx1), Linv_Kx2, [[-1], [-2]])
+        Kx1 = leading_transpose(Kx1, [..., -1, -2])  # [..., L, num_data, N]
+        Linv_Kx1 = tf.linalg.triangular_solve(L, Kx1)  # [..., L, num_data, N]
+        Linv_Kx2 = tf.linalg.triangular_solve(L, Kx2)  # [L, num_data, M]
+
+        # The line below is just A^T*B over the last 2 dimensions.
+        cov = K12 - tf.einsum("...lji,ljk->...lik", Linv_Kx1, Linv_Kx2)  # [..., L, N, M]
 
         tf.debugging.assert_shapes(
-            [(query_points_1, ["N", "D"]), (query_points_2, ["M", "D"]), (cov, ["N", "M"])]
+            [
+                (query_points_1, [..., "N", "D"]),
+                (query_points_2, ["M", "D"]),
+                (cov, [..., "L", "N", "M"]),
+            ]
         )
 
         return cov
@@ -211,6 +258,202 @@ class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel):
 
         multiple_assign(self.model, current_best_parameters)
 
+    def trajectory_sampler(self) -> TrajectorySampler:
+        """
+        Return a trajectory sampler. For :class:`GaussianProcessRegression`, we build
+        trajectories using a random Fourier feature approximation.
+
+        :return: The trajectory sampler.
+        """
+        models_data = Dataset(self.model.data[0].value(), self.model.data[1].value())
+        return RandomFourierFeatureTrajectorySampler(self, models_data, self._num_rff_features)
+
+    def conditional_predict_f(
+        self, query_points: TensorType, additional_data: Dataset
+    ) -> tuple[TensorType, TensorType]:
+        """
+        Returns the marginal GP distribution at query_points conditioned on both the model
+        and some additional data, using exact formula. See :cite:`chevalier2014corrected`
+        (eqs. 8-10) for details.
+
+        :param query_points: Set of query points with shape [M, D]
+        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
+                 with shape [..., N, L]
+        :return: mean_qp_new: predictive mean at query_points, with shape [..., M, L],
+                 and var_qp_new: predictive variance at query_points, with shape [..., M, L]
+        """
+
+        tf.debugging.assert_shapes(
+            [
+                (additional_data.query_points, [..., "N", "D"]),
+                (additional_data.observations, [..., "N", "L"]),
+                (query_points, ["M", "D"]),
+            ],
+            message="additional_data must have query_points with shape [..., N, D]"
+            " and observations with shape [..., N, L], and query_points "
+            "should have shape [M, D]",
+        )
+
+        if isinstance(self.model, SGPR):
+            raise NotImplementedError("Conditional predict f is not supported for SGPR.")
+
+        mean_add, cov_add = self.model.predict_f(
+            additional_data.query_points, full_cov=True
+        )  # [..., N, L], [..., L, N, N]
+        mean_qp, var_qp = self.model.predict_f(query_points, full_cov=False)  # [M, L], [M, L]
+
+        cov_cross = self.covariance_between_points(
+            additional_data.query_points, query_points
+        )  # [..., L, N, M]
+
+        cov_shape = tf.shape(cov_add)
+        noise = self.model.likelihood.variance * tf.eye(
+            cov_shape[-2], batch_shape=cov_shape[:-2], dtype=cov_add.dtype
+        )
+        L_add = tf.linalg.cholesky(cov_add + noise)  # [..., L, N, N]
+        A = tf.linalg.triangular_solve(L_add, cov_cross, lower=True)  # [..., L, N, M]
+        var_qp_new = var_qp - leading_transpose(
+            tf.reduce_sum(A ** 2, axis=-2), [..., -1, -2]
+        )  # [..., M, L]
+
+        mean_add_diff = additional_data.observations - mean_add  # [..., N, L]
+        mean_add_diff = leading_transpose(mean_add_diff, [..., -1, -2])[..., None]  # [..., L, N, 1]
+        AM = tf.linalg.triangular_solve(L_add, mean_add_diff)  # [..., L, N, 1]
+
+        mean_qp_new = mean_qp + leading_transpose(
+            (tf.matmul(A, AM, transpose_a=True)[..., 0]), [..., -1, -2]
+        )  # [..., M, L]
+
+        tf.debugging.assert_shapes(
+            [
+                (additional_data.observations, [..., "N", "L"]),
+                (query_points, ["M", "D"]),
+                (mean_qp_new, [..., "M", "L"]),
+                (var_qp_new, [..., "M", "L"]),
+            ],
+            message="received unexpected shapes computing conditional_predict_f,"
+            "check model kernel structure?",
+        )
+
+        return mean_qp_new, var_qp_new
+
+    def conditional_predict_joint(
+        self, query_points: TensorType, additional_data: Dataset
+    ) -> tuple[TensorType, TensorType]:
+        """
+        Predicts the joint GP distribution at query_points conditioned on both the model
+        and some additional data, using exact formula. See :cite:`chevalier2014corrected`
+        (eqs. 8-10) for details.
+
+        :param query_points: Set of query points with shape [M, D]
+        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
+                 with shape [..., N, L]
+        :return: mean_qp_new: predictive mean at query_points, with shape [..., M, L],
+                 and cov_qp_new: predictive covariance between query_points, with shape
+                 [..., L, M, M]
+        """
+
+        tf.debugging.assert_shapes(
+            [
+                (additional_data.query_points, [..., "N", "D"]),
+                (additional_data.observations, [..., "N", "L"]),
+                (query_points, ["M", "D"]),
+            ],
+            message="additional_data must have query_points with shape [..., N, D]"
+            " and observations with shape [..., N, L], and query_points "
+            "should have shape [M, D]",
+        )
+
+        if isinstance(self.model, SGPR):
+            raise NotImplementedError("Conditional predict f is not supported for SGPR.")
+
+        leading_dims = tf.shape(additional_data.query_points)[:-2]  # [...]
+        new_shape = tf.concat([leading_dims, tf.shape(query_points)], axis=0)  # [..., M, D]
+        query_points_r = tf.broadcast_to(query_points, new_shape)  # [..., M, D]
+        points = tf.concat([additional_data.query_points, query_points_r], axis=-2)  # [..., N+M, D]
+
+        mean, cov = self.model.predict_f(points, full_cov=True)  # [..., N+M, L], [..., L, N+M, N+M]
+
+        N = tf.shape(additional_data.query_points)[-2]
+
+        mean_add = mean[..., :N, :]  # [..., N, L]
+        mean_qp = mean[..., N:, :]  # [..., M, L]
+
+        cov_add = cov[..., :N, :N]  # [..., L, N, N]
+        cov_qp = cov[..., N:, N:]  # [..., L, M, M]
+        cov_cross = cov[..., :N, N:]  # [..., L, N, M]
+
+        cov_shape = tf.shape(cov_add)
+        noise = self.model.likelihood.variance * tf.eye(
+            cov_shape[-2], batch_shape=cov_shape[:-2], dtype=cov_add.dtype
+        )
+        L_add = tf.linalg.cholesky(cov_add + noise)  # [..., L, N, N]
+        A = tf.linalg.triangular_solve(L_add, cov_cross, lower=True)  # [..., L, N, M]
+        cov_qp_new = cov_qp - tf.matmul(A, A, transpose_a=True)  # [..., L, M, M]
+
+        mean_add_diff = additional_data.observations - mean_add  # [..., N, L]
+        mean_add_diff = leading_transpose(mean_add_diff, [..., -1, -2])[..., None]  # [..., L, N, 1]
+        AM = tf.linalg.triangular_solve(L_add, mean_add_diff)  # [..., L, N, 1]
+        mean_qp_new = mean_qp + leading_transpose(
+            (tf.matmul(A, AM, transpose_a=True)[..., 0]), [..., -1, -2]
+        )  # [..., M, L]
+
+        tf.debugging.assert_shapes(
+            [
+                (additional_data.observations, [..., "N", "L"]),
+                (query_points, ["M", "D"]),
+                (mean_qp_new, [..., "M", "L"]),
+                (cov_qp_new, [..., "L", "M", "M"]),
+            ],
+            message="received unexpected shapes computing conditional_predict_joint,"
+            "check model kernel structure?",
+        )
+
+        return mean_qp_new, cov_qp_new
+
+    def conditional_predict_f_sample(
+        self, query_points: TensorType, additional_data: Dataset, num_samples: int
+    ) -> TensorType:
+        """
+        Generates samples of the GP at query_points conditioned on both the model
+        and some additional data.
+
+        :param query_points: Set of query points with shape [M, D]
+        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
+                 with shape [..., N, L]
+        :param num_samples: number of samples
+        :return: samples of f at query points, with shape [..., num_samples, M, L]
+        """
+
+        if isinstance(self.model, SGPR):
+            raise NotImplementedError("Conditional predict y is not supported for SGPR.")
+
+        mean_new, cov_new = self.conditional_predict_joint(query_points, additional_data)
+        mean_for_sample = tf.linalg.adjoint(mean_new)  # [..., L, N]
+        samples = sample_mvn(
+            mean_for_sample, cov_new, full_cov=True, num_samples=num_samples
+        )  # [..., (S), P, N]
+        return tf.linalg.adjoint(samples)  # [..., (S), N, L]
+
+    def conditional_predict_y(
+        self, query_points: TensorType, additional_data: Dataset
+    ) -> tuple[TensorType, TensorType]:
+        """
+        Generates samples of y from the GP at query_points conditioned on both the model
+        and some additional data.
+
+        :param query_points: Set of query points with shape [M, D]
+        :param additional_data: Dataset with query_points with shape [..., N, D] and observations
+                 with shape [..., N, L]
+        :return: predictive variance at query_points, with shape [..., M, L],
+                 and predictive variance at query_points, with shape [..., M, L]
+        """
+
+        if isinstance(self.model, SGPR):
+            raise NotImplementedError("Conditional predict y is not supported for SGPR.")
+        f_mean, f_var = self.conditional_predict_f(query_points, additional_data)
+        return self.model.likelihood.predict_mean_and_var(f_mean, f_var)
+
 
 class NumDataPropertyMixin:
     """Mixin class for exposing num_data as a property, stored in a tf.Variable. This is to work
@@ -257,6 +500,8 @@ class SparseVariational(GPflowPredictor, TrainableProbabilisticModel):
         super().__init__(optimizer)
         self._model = model
 
+        check_optimizer(optimizer)
+
         # GPflow stores num_data as a number. However, since we want to be able to update it
         # without having to retrace the acquisition functions, put it in a Variable instead.
         # So that the elbo method doesn't fail we also need to turn it into a property.
@@ -266,7 +511,7 @@ class SparseVariational(GPflowPredictor, TrainableProbabilisticModel):
 
     def __repr__(self) -> str:
         """"""
-        return f"SparseVariational({self._model!r}, {self.optimizer!r})"
+        return f"SparseVariational({self.model!r}, {self.optimizer!r})"
 
     @property
     def model(self) -> SVGP:
@@ -341,21 +586,26 @@ class VariationalGaussianProcess(GPflowPredictor, TrainableProbabilisticModel):
             optimizer = BatchOptimizer(tf.optimizers.Adam(), batch_size=100)
 
         super().__init__(optimizer)
-        self._model = model
+
+        check_optimizer(self.optimizer)
 
         if use_natgrads:
-            if not isinstance(self._optimizer, BatchOptimizer) or not isinstance(
-                self._optimizer.optimizer, tf.optimizers.Optimizer
-            ):
+            if not isinstance(self.optimizer.optimizer, tf.optimizers.Optimizer):
                 raise ValueError(
                     f"""
                     Natgrads can only be used with a BatchOptimizer wrapper using an instance of
-                    tf.optimizers.Optimizer, however received f{self._optimizer}.
+                    tf.optimizers.Optimizer, however received {self.optimizer}.
                     """
                 )
-
             natgrad_gamma = 0.1 if natgrad_gamma is None else natgrad_gamma
         else:
+            if isinstance(self.optimizer.optimizer, tf.optimizers.Optimizer):
+                raise ValueError(
+                    f"""
+                    If not using natgrads an Optimizer wrapper should be used with
+                    gpflow.optimizers.Scipy, however received {self.optimizer}.
+                    """
+                )
             if natgrad_gamma is not None:
                 raise ValueError(
                     """
@@ -363,6 +613,7 @@ class VariationalGaussianProcess(GPflowPredictor, TrainableProbabilisticModel):
                     """
                 )
 
+        self._model = model
         self._use_natgrads = use_natgrads
         self._natgrad_gamma = natgrad_gamma
 
@@ -388,7 +639,7 @@ class VariationalGaussianProcess(GPflowPredictor, TrainableProbabilisticModel):
 
     def __repr__(self) -> str:
         """"""
-        return f"VariationalGaussianProcess({self._model!r}, {self.optimizer!r})"
+        return f"VariationalGaussianProcess({self.model!r}, {self.optimizer!r})"
 
     @property
     def model(self) -> VGP:
